import itertools
import logging
import uuid
from typing import Iterable, TypedDict, get_args

from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.db import models, transaction
from ordered_model.models import OrderedModel
from rich.console import Console

from formkit_ninja import formkit_schema

console = Console()
log = console.log

logger = logging.getLogger()


class UuidIdModel(models.Model):
    """
    Consistently use fields which will
    help with syncing data:
     - UUID field is the ID
     - Created field
     - Last Modified field
     - updated_by (optional)
     - created_by (optional)
    """

    class Meta:
        abstract = True

    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False, unique=True)
    created = models.DateTimeField(auto_now_add=True, blank=True, null=True)
    updated = models.DateTimeField(auto_now=True, blank=True, null=True)
    created_by = models.ForeignKey(
        settings.AUTH_USER_MODEL, on_delete=models.PROTECT, related_name="+", blank=True, null=True
    )
    updated_by = models.ForeignKey(
        settings.AUTH_USER_MODEL, on_delete=models.PROTECT, related_name="+", blank=True, null=True
    )


class OptionDict(TypedDict):
    value: str
    label: str


class OptionGroup(models.Model):
    """
    This intended to be a "collection" of choices
    For instance all the values in a single PNDS zTable
    Also intended to allow users to add / modify their __own__ 'Options'
    for idb and formkit to recognize
    """

    group = models.CharField(max_length=1024, primary_key=True, help_text="The label to use for these options")
    content_type = models.ForeignKey(
        ContentType,
        on_delete=models.PROTECT,
        null=True,
        blank=True,
        help_text="This is an optional reference to the original source object for this set of options (typically a table from which we copy options)",
    )

    def __str__(self):
        return f"{self.group}"

    @classmethod
    def copy_table(cls, model: models.Model, field: str, language: str | None = "en", group: str | None = None):
        """
        Copy an existing table of options into this OptionGroup
        """

        with transaction.atomic():
            group, group_created = cls.objects.get_or_create(
                group=group, content_type=ContentType.objects.get_for_model(model)
            )
            log(group)

            for obj in model.objects.values("pk", field):
                option, option_created = Option.objects.get_or_create(
                    object_id=obj["pk"],
                    group=group,
                    value=obj["pk"],
                )
                OptionLabel.objects.get_or_create(option=option, label=obj[field] or "", lang=language)


class Option(OrderedModel, UuidIdModel):
    """
    This is a key/value field representing one "option" for a FormKit property
    The translated values for this option are in the `Translatable` table
    """

    object_id = models.IntegerField(
        null=True,
        blank=True,
        help_text="This is a reference to the primary key of the original source object (typically a PNDS ztable ID) or a user-specified ID for a new group",
    )
    last_updated = models.DateTimeField(auto_now=True)
    group = models.ForeignKey(OptionGroup, on_delete=models.CASCADE)

    class Meta:
        constraints = [models.UniqueConstraint(fields=["group", "object_id"], name="unique_option_id")]
        ordering = (
            "group",
            "order",
        )

    value = models.CharField(max_length=1024)
    field = models.ForeignKey(
        "FormKitSchemaNode",
        on_delete=models.CASCADE,
        limit_choices_to={"node__formkit__in": ["select", "radio", "dropdown"]},
        null=True,
        blank=True,
        help_text="The ID of an Option node (select, dropdown or radio) if this option is embedded as part of a selection node",
    )
    order_with_respect_to = "group"

    @classmethod
    def from_pydantic(cls, options: list[str] | list[OptionDict]) -> Iterable["Option"]:
        for option in options:
            if isinstance(option, str):
                yield cls(value=option, label=option)
            elif isinstance(option, dict) and option.keys() == {"value", "label"}:
                yield cls(**option)

    def __str__(self):
        return f"{self.group.group}::{self.value}"


class OptionLabel(models.Model):
    option = models.ForeignKey("Option", on_delete=models.CASCADE)
    label = models.CharField(max_length=1024)
    lang = models.CharField(
        max_length=4, default="en", choices=(("en", "English"), ("tet", "Tetum"), ("pt", "Portugese"))
    )

    def save(self, *args, **kwargs):
        """
        When saved, save also my "option" so that its last_updated is set
        """
        if self.option is not None:
            self.option.save()
        return super().save(*args, **kwargs)

    class Meta:
        constraints = [models.UniqueConstraint(fields=["option", "lang"], name="unique_option_label")]


class FormComponents(OrderedModel, UuidIdModel):
    """
    A model relating "nodes" of a schema to a schema with model ordering
    """

    schema = models.ForeignKey("FormKitSchema", on_delete=models.CASCADE)
    # This is null=True so that a new FormComponent can be added from the admin inline
    node = models.ForeignKey("FormKitSchemaNode", on_delete=models.CASCADE, null=True, blank=True)
    label = models.CharField(max_length=1024, help_text="Used as a human-readable label", null=True, blank=True)

    order_with_respect_to = "schema"

    class Meta:
        ordering = ("schema", "order")

    def __str__(self):
        return f"{self.node}[{self.order}]: {self.schema}"


class Membership(OrderedModel, UuidIdModel):
    """
    This is an ordered m2m model representing
    how parts of a "FormKit group" are arranged
    """

    group = models.ForeignKey(
        "FormKitSchemaNode",
        on_delete=models.CASCADE,
        limit_choices_to={"node__formkit": "group"},
    )
    member = models.ForeignKey("FormKitSchemaNode", on_delete=models.CASCADE, related_name="members")
    order_with_respect_to = "group"


class NodeChildren(OrderedModel, UuidIdModel):
    """
    This is an ordered m2m model representing
    the "children" of an HTML element
    """

    parent = models.ForeignKey(
        "FormKitSchemaNode",
        on_delete=models.CASCADE,
    )
    child = models.ForeignKey("FormKitSchemaNode", on_delete=models.CASCADE, related_name="parent")
    order_with_respect_to = "parent"

    class Meta:
        ordering = ("order",)


class FormKitSchemaNode(UuidIdModel):
    """
    This represents a single "Node" in a FormKit schema.
    There are several different types of node which may be defined:
    FormKitSchemaDOMNode
    | FormKitSchemaComponent
    | FormKitSchemaTextNode
    | FormKitSchemaCondition
    | FormKitSchemaFormKit
    """

    NODE_TYPE_CHOICES = (
        ("$cmp", "Component"),
        ("text", "Text"),
        ("condition", "Condition"),
        ("$formkit", "FormKit"),
        ("$el", "Element"),
        ("raw", "Raw JSON"),
    )
    FORMKIT_CHOICES = [(t, t) for t in get_args(formkit_schema.FORMKIT_TYPE)]

    ELEMENT_TYPE_CHOICES = [("p", "p"), ("h1", "h1"), ("h2", "h2"), ("span", "span")]
    node_type = models.CharField(max_length=256, choices=NODE_TYPE_CHOICES, blank=True, help_text="")
    description = models.CharField(
        max_length=4000,
        null=True,
        blank=True,
        help_text="Decribe the type of data / reason for this component",
    )
    label = models.CharField(max_length=1024, help_text="Used as a human-readable label", null=True, blank=True)
    group = models.ManyToManyField("self", through=Membership, blank=True)
    children = models.ManyToManyField("self", through=NodeChildren, blank=True)

    node = models.JSONField(
        null=True,
        blank=True,
        help_text="A JSON representation of select parts of the FormKit schema",
    )

    additional_props = models.JSONField(
        null=True,
        blank=True,
        help_text="User space for additional, less used props",
    )

    text_content = models.TextField(
        null=True, blank=True, help_text="Content for a text element, for children of an $el type component"
    )

    def __str__(self):
        return f"{self.label}" if self.label else f"{self.node_type} {self.id}"

    def save(self, *args, **kwargs):
        """
        On save validate the 'node' field matches the 'FormKitNode'
        """
        return super().save(*args, **kwargs)

    @property
    def node_options(self):
        """
        Because "options" are translated and
        separately stored, this step is necessary to
        reinstate them
        """
        options: Iterable[Option] = self.option_set.all()
        return [{"value": option.value, "label": option.label} for option in options]

    def get_node_values(self) -> dict:
        """
        Reify a 'dict' instance suitable for creating
        a FormKit Schema node from
        """
        if not self.node:
            return {}
        values = {**self.node}

        # Options may come from a string in the node, or
        # may come from an m2m
        if self.node_options:
            values["options"] = self.node_options

        children = [c.get_node_values() for c in self.children.order_by("parent__order")]
        if children:
            values["children"] = children
        return values

    def get_node(self) -> formkit_schema.Node:
        """
        Return a "decorated" node instance
        with restored options and translated fields
        """
        node_content = self.get_node_values()
        if node_content == {}:
            if self.node_type == "$el":
                node_content = {"$el": "span"}
            elif self.node_type == "$formkit":
                node_content = {"$formkit": "text"}
        formkit_node = formkit_schema.FormKitNode.parse_obj(node_content)
        return formkit_node.__root__

    @classmethod
    def from_pydantic(
        cls, input_models: formkit_schema.FormKitSchemaProps | Iterable[formkit_schema.FormKitSchemaProps]
    ) -> Iterable[tuple["FormKitSchemaNode", Iterable[Option]]]:
        if isinstance(input_models, Iterable) and not isinstance(input_models, formkit_schema.FormKitSchemaProps):
            yield from (cls.from_pydantic(n) for n in input_models)

        elif isinstance(input_models, formkit_schema.FormKitSchemaProps):
            input_model = input_models
            instance = cls()
            log(f"[green]Creating {instance}")
            if options := getattr(input_model, "options", None):
                if isinstance(options, str):
                    option_models = Option.objects.none()
                else:
                    option_models = Option.from_pydantic(options)
                    instance.save()
                    for o in option_models:
                        o.field = instance
                        log(f"[green]    Adding option {o} to {instance}")
                        o.save()
            else:
                option_models = Option.objects.none()
            if label := getattr(input_model, "html_id", None):
                instance.label = label
            instance.node = input_model.dict(
                exclude={"options", "children", "additional_props"}, exclude_none=True, exclude_unset=True
            )

            # Node types
            if props := getattr(input_model, "additional_props", None):
                instance.additional_props = props

            node_type = getattr(input_model, "node_type")
            if node_type == "condition":
                instance.node_type = "condition"
            elif node_type == "formkit":
                instance.node_type = "$formkit"
            elif node_type == "element":
                instance.node_type = "$el"
            elif node_type == "component":
                instance.node_type = "$cmp"

            child_nodes = getattr(input_model, "children")
            # All other properties are passed directly
            log(f"[green]Yielding: {instance}")

            # Add the "options" if it is a 'text' type getter
            if options := getattr(input_model, "options", None):
                if isinstance(options, str):
                    instance.node["options"] = options

            yield instance, option_models

            if child_nodes:
                instance.save()
                instance.refresh_from_db()
                for child_node in child_nodes:
                    for c, c_opts in cls.from_pydantic(child_node):
                        log(f"[green]    Adding child node {c} to {instance}")
                        c.save()
                        NodeChildren.objects.create(parent=instance, child=c)

        else:
            raise TypeError(f"Expected FormKitNode or Iterable[FormKitNode], got {type(input_models)}")

    def to_pydantic(self):
        return formkit_schema.FormKitNode.parse_obj(self.get_node_values())


class SchemaManager(models.Manager):
    """
    Provides prefetching which we'll almost always want to have
    """

    def get_queryset(self):
        return super().get_queryset().prefetch_related("nodes", "nodes__children")


class FormKitSchema(UuidIdModel):
    """
    This represents a "FormKitSchema" which is an heterogenous
    collection of items.
    """

    label = models.TextField(null=True, blank=True, help_text="Used as a human-readable label")
    nodes = models.ManyToManyField(FormKitSchemaNode, through=FormComponents)
    objects = SchemaManager()

    def get_schema_values(self):
        """
        Return a list of "node" dicts
        """
        nodes: Iterable[FormKitSchemaNode] = self.nodes.all()
        for node in nodes:
            yield node.get_node_values()

    def to_pydantic(self):
        values = list(self.get_schema_values())
        return formkit_schema.FormKitSchema.parse_obj(values)

    def __str__(self):
        return f"{self.label}" or f"{str(self.id)[:8]}"

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)

    @classmethod
    def from_pydantic(cls, input_model: formkit_schema.FormKitSchema) -> "FormKitSchema":
        """
        Converts a given Pydantic representation of a Schema
        to Django database fields
        """
        instance = cls.objects.create()
        for node, options in itertools.chain.from_iterable(FormKitSchemaNode.from_pydantic(input_model.__root__)):
            log(f"[yellow]Saving {node}")
            node.save()
            if options:
                for option in options:
                    option.field = node
                    option.save()
            FormComponents.objects.create(schema=instance, node=node)
        logger.info("Schema load from JSON done")
        return instance

    @classmethod
    def from_json(cls, input_file: dict):
        """
        Converts a given JSON string to a suitable
        Django representation
        """
        schema_instance = formkit_schema.FormKitSchema.parse_obj(input_file)
        return cls.from_pydantic(schema_instance)



# Generated by Django 4.2.1 on 2023-08-28 05:07

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("formkit_ninja", "0003_delete_translatable"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="formkitschema",
            name="key",
        ),
        migrations.RemoveField(
            model_name="formkitschemanode",
            name="translation_context",
        ),
        migrations.AddField(
            model_name="formkitschema",
            name="label",
            field=models.TextField(blank=True, help_text="Used as a human-readable label", null=True),
        ),
        migrations.AlterField(
            model_name="formcomponents",
            name="label",
            field=models.CharField(blank=True, help_text="Used as a human-readable label", max_length=1024, null=True),
        ),
        migrations.AlterField(
            model_name="formcomponents",
            name="node",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="formkit_ninja.formkitschemanode",
            ),
        ),
        migrations.AlterField(
            model_name="formkitschemanode",
            name="label",
            field=models.CharField(blank=True, help_text="Used as a human-readable label", max_length=1024, null=True),
        ),
    ]



from formkit_ninja import models


def create_location_nodes():
    """
    Yield the nodes used by location schemas
    """

    district, c = models.FormKitSchemaNode.objects.update_or_create(
        label="municipality",
        defaults=dict(
            node_type="$formkit",
            node=dict(
                label="Municipality",
                html_id="municipality",
                name="municipality",
                key="municipality",
                formkit="select",
                options="$getLocations()",
            ),
        ),
    )
    yield district, c

    admin_post, c = models.FormKitSchemaNode.objects.get_or_create(
        label="admin_post",
        defaults=dict(
            node_type="$formkit",
            node=dict(
                label="Administrative Post",
                html_id="admin_post",
                name="admin_post",
                key="admin_post",
                formkit="select",
                if_condition='$getLocations($get("municipality").value)',
                options='$getLocations($get("municipality").value)',
            ),
        ),
    )
    yield admin_post, c

    suco, c = models.FormKitSchemaNode.objects.get_or_create(
        label="suco",
        defaults=dict(
            node_type="$formkit",
            node=dict(
                label="Suco",
                html_id="suco",
                name="suco",
                key="suco",
                formkit="select",
                if_condition='$getLocations($get("admin_post").value)',
                options='$getLocations($get("municipality").value, $get("admin_post").value)',
            ),
        ),
    )
    yield suco, c

    aldeia, c = models.FormKitSchemaNode.objects.get_or_create(
        label="aldeia",
        defaults=dict(
            node_type="$formkit",
            node=dict(
                label="Aldeia",
                html_id="aldeia",
                name="aldeia",
                key="aldeia",
                formkit="select",
                if_condition='$getLocations($get("suco").value)',
                options='$getLocations($get("municipality").value, $get("admin_post").value, $get("suco").value)',
            ),
        ),
    )
    yield aldeia, c


def add_location_schema():
    """
    This is a "general" location picker without the special SF 1.1 logic
    """

    # The SF11 schema is represented by these parts
    locations, created = models.FormKitSchema.objects.get_or_create(key="Locations except SF 1.1")

    yield locations, created

    # Locations has: suco, aldeia, postu_admin, district nodes
    # From the admin, you'll set the "admin key" and the "node type"

    municipality, admin_post, suco, aldeia = (node for (node, _) in create_location_nodes())

    yield models.FormComponents.objects.update_or_create(
        defaults=dict(
            schema=locations,
            node=municipality,
        ),
        key="Municipalities for the Location Select schema",
    )

    yield models.FormComponents.objects.update_or_create(
        key="Administrative Posts for the Location Select schema",
        defaults=dict(
            schema=locations,
            node=admin_post,
        ),
    )

    yield models.FormComponents.objects.update_or_create(
        defaults=dict(
            schema=locations,
            node=suco,
        ),
        key="Suco for the Location Select schema",
    )

    yield models.FormComponents.objects.update_or_create(
        key="Aldeias for the Location Select schema",
        defaults=dict(
            schema=locations,
            node=aldeia,
        ),
    )



# Generated by Django 4.2.1 on 2023-08-28 05:39

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("formkit_ninja", "0007_alter_option_field"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="option",
            name="label",
        ),
    ]



# Generated by Django 4.2.1 on 2023-08-28 05:26

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("formkit_ninja", "0005_option_last_updated_option_object_id_optionlabel_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="option",
            name="group",
            field=models.ForeignKey(
                default="", on_delete=django.db.models.deletion.CASCADE, to="formkit_ninja.optiongroup"
            ),
            preserve_default=False,
        ),
        migrations.AddConstraint(
            model_name="option",
            constraint=models.UniqueConstraint(fields=("group", "object_id"), name="unique_option_id"),
        ),
    ]



from django.core.management.base import BaseCommand

from formkit_ninja.management.commands.common_nodes import create_location_nodes


def create():
    # The SF 13 form has the following schemas
    # meetingInformation
    # participants
    # projectteam
    # planning
    # sukus

    # The SF 13 form has the following components
    # For 'meeting information':
    # municipality
    # adminpost
    # sucos
    # date

    # For 'partisipants':
    # suku_comittee_member_female
    # suku_committee_member_male
    # community_member_female
    # community_member_male
    # participants_with_disability_female
    # participants_with_disability_male

    # For 'projectteam':
    # A 'repeater' (ooh, fancy!)
    # with lots of nested children, aargh
    ...


class Command(BaseCommand):
    help = """
        Create the Partisipa 'SF13' form
        as a set of Django models / FormKit integration
    """

    def handle(self, *args, **options):
        for instance, created in create_location_nodes():
            if options["verbosity"] > 1:
                self.stdout.write(self.style.SUCCESS(f"{'Created' if created else 'Updated'} {instance}"))
        for instance, created in create():
            if options["verbosity"] > 1:
                self.stdout.write(self.style.SUCCESS(f"{'Created' if created else 'Updated'} {instance}"))



from __future__ import annotations

import logging
import operator
import warnings
from functools import reduce
from typing import Any, Optional

from django import forms
from django.contrib import admin
from django.http import HttpRequest
from ordered_model.admin import OrderedInlineModelAdminMixin, OrderedModelAdmin, OrderedTabularInline

from formkit_ninja import formkit_schema, models

logger = logging.getLogger(__name__)


class ItemAdmin(OrderedModelAdmin):
    list_display = ("name", "move_up_down_links")


class JsonDecoratedFormBase(forms.ModelForm):
    """
    Adds additional fields to the admin where a model has a JSON field
    and some appropriate (tbc?) field parameters
    """

    # extra = forms.CharField(label="Extra", max_length=128, required=False)
    # hello_world = forms.CharField(widget=forms.NumberInput, required=False)

    # key is the name of a `models.JSONField` on the model
    # value is a list of fields to get/set in that JSON field
    _json_fields: dict[str, tuple[str]] = {"my_json_field": ("formkit", "description", "name", "key", "html_id")}

    def get_json_fields(self) -> dict[str, tuple[str]]:
        """
        Custom which json fields will be included in the return
        """
        return self._json_fields

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        instance = kwargs["instance"]
        for field, keys in self.get_json_fields().items():
            # Extract the dict of JSON values from the model instance if supplied
            values = {}
            if instance:
                values = getattr(instance, field, {}) or {}  # Don't allow none
            for key in keys:
                # The value, extracted from the JSON value in the database
                field_value = values.get(key, None)
                # The initial value of the admin form is set to the value of the JSON attrib
                if key in self.fields:
                    self.fields.get(key).initial = field_value
                else:
                    warnings.warn(f"Unassociated JSON field found: {key}")
        return

    def save(self, commit=True):
        """
        Updates the JSON field(s) from the fields specified in the `_json_fields` dict
        """

        for field, keys in self.get_json_fields().items():
            # Populate a JSON field in a model named "form"
            # from a set of standard form elements
            data = {}
            for key in keys:
                if cleaned_data := self.cleaned_data[key]:
                    data[key] = cleaned_data
        setattr(self.instance, field, data)
        return super().save(commit=commit)


class NewFormKitForm(forms.ModelForm):
    class Meta:
        model = models.FormKitSchemaNode
        fields = ("label", "node_type", "description")


class OptionForm(forms.ModelForm):
    class Meta:
        model = models.Option
        exclude = ()


class FormComponentsForm(forms.ModelForm):
    class Meta:
        model = models.FormComponents
        exclude = ()


class FormKitSchemaNodeOptionsInline(OrderedTabularInline):
    model = models.Option
    form = OptionForm
    fields = ("label", "value", "id")
    readonly_fields = (
        # "id",
        "order",
        "move_up_down_links",
    )
    ordering = ("order",)
    extra = 0


class FormKitSchemaComponentInline(OrderedTabularInline):
    model = models.FormComponents
    readonly_fields = (
        "node",
        "created_by",
        "updated_by",
        "order",
        "move_up_down_links",
    )
    ordering = ("order",)
    extra = 0


class FormKitNodeGroupForm(JsonDecoratedFormBase):
    class Meta:
        model = models.FormKitSchemaNode
        fields = (
            "label",
            "description",
        )

    _json_fields = {
        "node": (
            "formkit",
            "if_condition",
        )
    }

    formkit = forms.ChoiceField(required=False, choices=models.FormKitSchemaNode.FORMKIT_CHOICES, disabled=True)
    if_condition = forms.CharField(
        widget=forms.TextInput,
        required=False,
    )


class FormKitNodeForm(JsonDecoratedFormBase):
    """
    This is the most common component type: the 'FormKit' schema node
    """

    class Meta:
        model = models.FormKitSchemaNode
        fields = ("label", "description", "additional_props")

    # The `_json_fields["node"]` item affects the admin form,
    # adding the fields included in the `FormKitSchemaProps.__fields__.items` dict
    _json_fields = {
        "node": (
            "formkit",
            "description",
            "name",
            "key",
            "html_id",
            "if_condition",
            "options",
            "label",
            "placeholder",
            "help",
            # Validation of fields
            "validation",
            "validationLabel",
            "validationVisibility",
            "validationMessages",
            "prefixIcon",
            # Numeric forms
            "min",
            "max",
            "step",
        )
    }

    formkit = forms.ChoiceField(required=False, choices=models.FormKitSchemaNode.FORMKIT_CHOICES)
    name = forms.CharField(required=False)
    if_condition = forms.CharField(widget=forms.TextInput, required=False)
    key = forms.CharField(required=False)
    label = forms.CharField(required=False)
    placeholder = forms.CharField(required=False)
    help = forms.CharField(required=False)
    html_id = forms.CharField(
        required=False, help_text="Use this ID if adding conditions to other fields (hint: $get(my_field).value === 8)"
    )
    options = forms.CharField(
        required=False, help_text="Use this if adding Options using a JS function (hint: $get(my_field).value )"
    )
    validation = forms.CharField(required=False)
    validationLabel = forms.CharField(required=False)
    validationVisibility = forms.CharField(required=False)
    validationMessages = forms.JSONField(required=False)
    validationRules = forms.CharField(
        required=False, help_text="A function for validation passed into the schema: a key on `formSchemaData`"
    )
    prefixIcon = forms.CharField(required=False)

    # NumberNode props

    max = forms.IntegerField(required=False)
    min = forms.IntegerField(required=False)
    step = forms.IntegerField(required=False)

    def get_fields(self, request, obj: models.FormKitSchemaNode):
        """
        Customise the returned fields based on the type
        of formkit node
        """
        return super().get_fields(request, obj)


class FormKitNodeRepeaterForm(FormKitNodeForm):
    def get_json_fields(self) -> dict[str, tuple[str]]:
        return {
            "node": (
                *(super()._json_fields["node"]),
                "addLabel",
                "upControl",
                "downControl",
                "itemsClass",
                "itemClass",
            )
        }

    addLabel = forms.CharField(required=False)
    upControl = forms.BooleanField(required=False)
    downControl = forms.BooleanField(required=False)
    itemsClass = forms.CharField(required=False)
    itemClass = forms.CharField(required=False)
    max = forms.IntegerField(required=False)
    min = forms.IntegerField(required=False)


class FormKitTextNode(JsonDecoratedFormBase):
    class Meta:
        model = models.FormKitSchemaNode
        fields = (
            "label",
            "description",
        )

    _json_fields = {"node": ("content",)}
    content = forms.CharField(
        widget=forms.TextInput,
        required=True,
    )


class FormKitElementForm(JsonDecoratedFormBase):
    class Meta:
        model = models.FormKitSchemaNode
        fields = ("label", "description", "text_content")

    _skip_translations = {"label", "placeholder"}
    _json_fields = {"node": ("el", "name", "if_condition", "classes")}

    el = forms.ChoiceField(required=False, choices=models.FormKitSchemaNode.ELEMENT_TYPE_CHOICES)
    name = forms.CharField(
        required=False,
    )
    classes = forms.CharField(
        required=False,
    )
    if_condition = forms.CharField(
        widget=forms.TextInput,
        required=False,
    )


class FormKitConditionForm(JsonDecoratedFormBase):
    class Meta:
        model = models.FormKitSchemaNode
        # fields = '__all__'
        fields = (
            "label",
            "description",
        )

    _json_fields = {"node": ("if_condition", "then_condition", "else_condition")}

    if_condition = forms.CharField(
        widget=forms.TextInput,
        required=False,
    )
    then_condition = forms.CharField(
        max_length=256,
        required=False,
    )
    else_condition = forms.CharField(
        max_length=256,
        required=False,
    )


class FormKitComponentForm(JsonDecoratedFormBase):
    class Meta:
        model = models.FormKitSchemaNode
        fields = (
            "label",
            "description",
        )

    _json_fields = {"node": ("if_condition", "then_condition", "else_condition")}


class MembershipInline(OrderedTabularInline):
    model = models.Membership
    fk_name = "group"
    extra = 0
    readonly_fields = (
        "order",
        "move_up_down_links",
    )
    ordering = ("order",)


class MembershipComponentInline(admin.StackedInline):
    model = models.Membership
    fk_name = "member"
    extra = 0


class NodeChildrenInline(OrderedTabularInline):
    """
    Nested HTML elements
    """

    model = models.NodeChildren
    fields = ("child", "id", "order", "move_up_down_links")
    readonly_fields = (
        "order",
        "move_up_down_links",
    )
    ordering = ("order",)
    fk_name = "parent"
    extra = 0


class FormKitSchemaForm(forms.ModelForm):
    class Meta:
        model = models.FormKitSchema
        exclude = ("name",)


@admin.register(models.FormKitSchemaNode)
class FormKitSchemaNodeAdmin(OrderedInlineModelAdminMixin, admin.ModelAdmin):
    list_display = ("label", "id", "node_type")

    def get_inlines(self, request, obj: models.FormKitSchemaNode | None):
        if not obj:
            return []

        formkit_node_type = (obj.node or {}).get("formkit", None)

        if formkit_node_type == "group":
            return [
                NodeChildrenInline,
            ]
        elif formkit_node_type in {"radio", "select", "dropdown"}:
            return [
                MembershipComponentInline,
                FormKitSchemaNodeOptionsInline,
            ]
        elif formkit_node_type:
            return [
                MembershipComponentInline,
            ]

        if obj.node_type == "$el":
            return [
                NodeChildrenInline,
            ]

        else:
            return []

    # # Note that although overridden these are necessary
    # inlines = [NodeChildrenInline]
    inlines = [FormKitSchemaNodeOptionsInline, NodeChildrenInline, MembershipInline]

    def get_fieldsets(
        self, request: HttpRequest, obj: Optional[models.FormKitSchemaNode] = ...
    ) -> list[tuple[Optional[str], dict[str, Any]]]:
        """
        Set fieldsets to control the layout of admin “add” and “change” pages.
        fieldsets is a list of two-tuples, in which each two-tuple represents a <fieldset>
        on the admin form page. (A <fieldset> is a “section” of the form.)
        """
        fieldsets: list[tuple[str, dict]] = []
        if not getattr(obj, "node_type", None):
            warnings.warn("Expected a 'Node' with a 'NodeType' in the admin form")

        if not obj:
            # This default form is returned before a field
            # type is selected
            return super().get_fieldsets(request, obj)
        try:
            node = obj.get_node()
        except Exception as E:
            warnings.warn(f"{E}")
            return fieldsets
        if isinstance(node, formkit_schema.FormKitSchemaProps) and not isinstance(
            node, (formkit_schema.GroupNode, formkit_schema.FormKitSchemaDOMNode, formkit_schema.RepeaterNode)
        ):
            # Field validation applies to formkit
            # except repeater and group
            fieldsets.append(
                (
                    "Field Validation",
                    {
                        "fields": (
                            "validation",
                            "validationLabel",
                            "validationVisibility",
                            "validationMessages",
                            "validationRules",
                        )
                    },
                )
            )
        elif isinstance(node, formkit_schema.RepeaterNode):
            if obj.node and obj.node.get("formkit", None) == "repeater":
                fieldsets.append(
                    (
                        "Repeater field properties",
                        {"fields": ("addLabel", "upControl", "downControl", "itemsClass", "itemClass")},
                    )
                )

        grouped_fields = reduce(operator.or_, (set(opts["fields"]) for _, opts in fieldsets), set())
        # Add 'ungrouped' fields
        fieldsets.insert(
            0, (None, {"fields": [field for field in self.get_fields(request, obj) if field not in grouped_fields]})
        )
        logger.info(fieldsets)
        return fieldsets

    def get_form(
        self,
        request: Any,
        obj: models.FormKitSchemaNode | None,
        change: None = None,
        **kwargs,
    ):
        if not obj:
            return NewFormKitForm
        try:
            node = obj.get_node()
        except Exception as E:
            warnings.warn(f"{E}")
            return NewFormKitForm
        if isinstance(node, formkit_schema.GroupNode):
            return FormKitNodeGroupForm
        elif isinstance(node, formkit_schema.RepeaterNode):
            return FormKitNodeRepeaterForm
        elif isinstance(node, formkit_schema.FormKitSchemaDOMNode):
            return FormKitElementForm
        elif isinstance(node, formkit_schema.FormKitSchemaComponent):
            return FormKitComponentForm
        elif isinstance(node, formkit_schema.FormKitSchemaCondition):
            return FormKitConditionForm
        elif isinstance(node, formkit_schema.FormKitSchemaProps):
            return FormKitNodeForm

        else:
            warnings.warn(f"Unable to determine form type for {obj}")
            return super().get_form(request, obj, change, **kwargs)


@admin.register(models.FormKitSchema)
class FormKitSchemaAdmin(OrderedInlineModelAdminMixin, admin.ModelAdmin):
    form = FormKitSchemaForm

    def get_inlines(self, request, obj: models.FormKitSchema | None):
        """
        For a "new object" do not show the Form Components
        """
        if not obj:
            return []
        return [
            FormKitSchemaComponentInline,
        ]

    inlines = [
        FormKitSchemaComponentInline,
    ]


@admin.register(models.FormComponents)
class FormComponentsAdmin(OrderedModelAdmin):
    list_display = ("label", "schema", "node", "move_up_down_links")


class OptionLabelInline(admin.TabularInline):
    model = models.OptionLabel
    extra = 0


class OptionInline(admin.TabularInline):
    model = models.Option
    extra = 0
    fields = ("group", "object_id", "value", "field")
    readonly_fields = ("group", "object_id", "value", "field")


@admin.register(models.Option)
class OptionAdmin(OrderedModelAdmin):
    list_display = ("object_id", "value", "field", "order", "group", "move_up_down_links")
    inlines = [OptionLabelInline]
    list_select_related = ("group",)
    readonly_fields = ("group", "object_id", "value", "field", "created_by", "updated_by")


@admin.register(models.OptionGroup)
class OptionGroupAdmin(admin.ModelAdmin):
    inlines = [OptionInline]
    ...


@admin.register(models.OptionLabel)
class OptionLabelAdmin(admin.ModelAdmin):
    list_display = (
        "label",
        "lang",
    )
    readonly_fields = ("option",)
    search_fields = ("label",)



# Generated by Django 4.2.1 on 2023-08-28 05:38

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("formkit_ninja", "0006_option_group_option_unique_option_id"),
    ]

    operations = [
        migrations.AlterField(
            model_name="option",
            name="field",
            field=models.ForeignKey(
                blank=True,
                help_text="The ID of an Option node (select, dropdown or radio) if this option is embedded as part of a selection node",
                limit_choices_to={"node__formkit__in": ["select", "radio", "dropdown"]},
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="formkit_ninja.formkitschemanode",
            ),
        ),
    ]



# Generated by Django 4.1.5 on 2023-05-09 05:37

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("formkit_ninja", "0002_alter_formcomponents_label_alter_formkitschema_key"),
    ]

    operations = [
        migrations.DeleteModel(
            name="Translatable",
        ),
    ]



import datetime
import warnings
from typing import List
from uuid import UUID

from django.db.models import F, Q
from django.http import HttpRequest, HttpResponse
from django.shortcuts import get_object_or_404
from ninja import ModelSchema, Router

from formkit_ninja import formkit_schema, models

router = Router(tags=["FormKit"])


class FormKitSchemaIn(ModelSchema):
    class Config:
        model = models.FormKitSchema
        model_fields = "__all__"


class FormKitSchemaListOut(ModelSchema):
    class Config:
        model = models.FormKitSchema
        model_fields = ("id", "label")


class FormComponentsOut(ModelSchema):
    node_id: UUID
    schema_id: UUID

    class Config:
        model = models.FormComponents
        model_fields = ("label",)


class NodeChildrenOut(ModelSchema):
    class Config:
        model = models.NodeChildren
        model_fields = ("parent", "child", "order")


class OptionLabel(ModelSchema):
    class Config:
        model = models.OptionLabel
        model_fields = ("lang", "label")


class Option(ModelSchema):
    group_name: str  # This is annotation of the model `content_type_model`
    optionlabel_set: list[OptionLabel]

    class Config:
        model = models.Option
        model_fields = ("id", "object_id")


@router.get("list-schemas", response=list[FormKitSchemaListOut])
def get_list_schemas(request):
    return models.FormKitSchema.objects.all()


@router.get("list-nodes", response=dict[str, formkit_schema.Node], exclude_defaults=True, exclude_none=True)
def get_formkit_nodes(request):
    """
    Get all of the FormKit nodes in the database
    """
    response: dict[str, formkit_schema.Node] = {}
    for node in models.FormKitSchemaNode.objects.all():
        try:
            response[f"{node.pk}"] = node.get_node()
        except:  # noqa: E722
            warnings.warn(f"An unparseable FormKit node was hit at {node.pk}")
    return response


@router.get("list-related-nodes", response=list[NodeChildrenOut], exclude_defaults=True, exclude_none=True)
def get_related_nodes(request):
    """
    Get all of the FormKit nodes in the database
    """
    return models.NodeChildren.objects.all()


@router.get(
    "list-components", response=List[FormComponentsOut], exclude_defaults=True, exclude_none=True, by_alias=True
)
def get_components(request):
    values = models.FormComponents.objects.all()
    return values


@router.get(
    "schema/{schema_id}",
    response=formkit_schema.FormKitSchema,
    exclude_none=True,
    by_alias=True,
)
def get_schemas(request, schema_id: UUID):
    schema: models.FormKitSchema = get_object_or_404(models.FormKitSchema.objects, id=schema_id)
    model = schema.to_pydantic()
    return model


@router.get(
    "node/{node_id}",
    response=formkit_schema.FormKitNode,
    exclude_none=True,
    by_alias=True,
)
def get_node(request, node_id: UUID):
    node: models.FormKitSchemaNode = get_object_or_404(models.FormKitSchemaNode.objects, id=node_id)
    instance = node.get_node()
    return instance


@router.get("/options", response=list[Option], exclude_none=True)
def list_options(request: HttpRequest, response: HttpResponse, since: str = None):
    """
    List all available options from the zTables
    and the associated Translations
    """

    model = Option.Config.model
    qs = model.objects.all()

    if since is not None and since != "":
        # Python 3.10 does not support a trailing 'Z'
        if since.endswith("Z"):
            since = since[:-1] + "+00:00"
        try:
            ts = datetime.datetime.fromisoformat(since)
            qs = qs.filter(Q(last_updated__gt=ts))
        except Exception as E:
            warnings.warn(f"{E}")
    try:
        response["X-lastupdated"] = qs.latest("last_updated").last_updated.isoformat()
    except model.DoesNotExist:
        # if there are no changes, use the same header as was sent
        if since:
            response["X-lastupdated"] = since
    return qs.annotate(group_name=F("group__group")).prefetch_related("optionlabel_set")



from __future__ import annotations

import warnings

# from collections import UserDict
from typing import Sequence

from django.conf import settings
from django.db.models import JSONField
from django.utils import translation


class WhitelistedKeysDict(dict):
    """
    Allow only certain keys.
    >>> t = WhitelistedKeysDict("ahi!", default_key="tet")
    >>> t["tet"]
    'ahi!'
    >>> t = WhitelistedKeysDict("fire!", default_key="en")
    >>> t["en"]
    'fire!'
    >>> t = WhitelistedKeysDict({"en": "fire!", "tet": "ahi!"}, permitted_keys={"tet", "en"})
    >>> t["tet"]
    'ahi!'
    >>> t["en"]
    'fire!'
    >>> set(t.keys()) == {'tet', 'en'}
    True
    >>> t = WhitelistedKeysDict({"en": "fire!", "tet": "ahi!", "noexist": "should-warn"}, permitted_keys={"tet", "en"})
    >>> "noexist" in t
    False
    >>> set(t.keys()) == {"tet", "en"}
    True
    """

    default_key: str | None
    permitted_keys: set[str]

    def __init__(
        self,
        dict_=None,
        /,
        permitted_keys: set[str] = set(),
        default_key: str | None = None,
        **kwargs,
    ):
        self.default_key = default_key
        self.permitted_keys = permitted_keys
        if default_key:
            self.permitted_keys.add(default_key)
        self.data = {}
        if isinstance(dict_, str):
            if self.default_key:
                return self.__init__({self.default_key: dict_})
            warnings.warn("No default key was set. You must have a default_key to initialize with a string.")
            return super().__init__()
        return super().__init__(dict_ or {})

    def __setitem__(self, key: str, item: str) -> None:
        if key in self.permitted_keys:
            return super().__setitem__(key, item)
        warnings.warn(f"ignored key not in whitelist: {key}")


class TranslatedValues(WhitelistedKeysDict):
    """
    dict rejects keys not in Django's LANGUAGES
    >>> t = TranslatedValues("ahi!")
    >>> t["tet"]
    'ahi!'
    >>> t = TranslatedValues({"en": "fire!", "tet": "ahi!"})
    >>> t["tet"]
    'ahi!'
    >>> t["en"]
    'fire!'
    >>> set(t.keys()) == {'tet', 'en'}
    True
    >>> t = TranslatedValues({"en": "fire!", "tet": "ahi!", "noexist": "should-warn"})
    >>> "noexist" in t
    False
    >>> set(t.keys()) == {"tet", "en"}
    True
    >>> t.value  # Note: only when in Django, if Django language is set to 'tet'
    'ahi!'
    """

    def __init__(self, dict_, /, **kwargs):
        return super().__init__(
            dict_,
            permitted_keys=set((lang[0] for lang in getattr(settings, "LANGUAGES", ()))),
            default_key=translation.get_language(),
            **kwargs,
        )

    @property
    def value(self):
        """
        Return the value at the current language, or fallback
        """
        return TranslatedValues.get_str(self)

    @staticmethod
    def get_str(
        dict_: TranslatedValues,
        lang: str = translation.get_language(),
        fallback: Sequence[str] = ("en",),
    ):
        """
        Return the value at the current language, or fallback (static method)
        """
        # A newly created value might error out here if the `__str__` ref's the value
        # hint the user that they may wish to `refresh_from_db`
        if dict_ is None:
            return ""

        if isinstance(dict_, str):
            warnings.warn(
                "TranslatedValues received a str, you may have a new model which hasn't called `refresh_from_db` yet"
            )
            return dict_
        if len(list(dict_)) == 0:
            return "empty"
        for lc in (lang, *fallback, list(dict_)[0]):
            if lc in dict_:
                return dict_[lc]


class TranslatedField(JSONField):
    """
    When a string is passed to this field
    it will be saved as a value corresponding
    to the "LANGUAGE_CODE" key
    """

    description = "A translated char field storing content as JSON"

    def from_db_value(self, value, expression, connection):
        value_: dict | None = super().from_db_value(value, expression, connection)
        return TranslatedValues(value_)

    def to_python(self, value):
        value_ = super().to_python(value)
        return TranslatedValues(value_)


if __name__ == "__main__":
    import doctest

    doctest.testmod()



from __future__ import annotations

import logging
import warnings
from html.parser import HTMLParser
from typing import Annotated, Any, ForwardRef, List, Literal, Type, TypedDict, TypeVar, Union

from pydantic import BaseModel, Field

"""
This is a port of selected parts of the FormKit schema
to Pydantic models.
"""


logger = logging.getLogger(__name__)

HtmlAttrs = dict[str, str | dict[str, str]]


Node = ForwardRef("Node")


class FormKitSchemaCondition(BaseModel):
    node_type: Literal["condition"]
    if_condition: str = Field(..., alias="if")
    then_condition: Node | List[Node] = Field(..., alias="then")
    else_condition: Node | List[Node] | None = Field(None, alias="else")


class FormKitSchemaConditionNoCircularRefs(BaseModel):
    """
    This class is defined in order to break circular references.
    Circular references (ie the `Node` in the then_condition) cause
    django_ninja to crash the server hard
    """

    node_type: Literal["condition"]
    if_condition: str = Field(..., alias="if")
    then_condition: str | list[str] = Field(..., alias="then")
    else_condition: str | list[str] | None = Field(None, alias="else")


class FormKitSchemaMeta(BaseModel):
    __root__: dict[str, str | float | int | bool | None]


class FormKitTypeDefinition(BaseModel):
    ...


class FormKitContextShape(BaseModel):
    type: Literal["input", "list", "group"]
    value: Any
    _value: Any


class FormKitListValue(BaseModel):
    __root__: str | list[str] | list[dict[str, str]]


class FormKitListStatement(BaseModel):
    """
    A full loop statement in tuple syntax. Can be read like "foreach value, key? in list"
    A 2 or 2 element tuple of value, key, and list or value, list
    """

    __root__: tuple[str, float | int | str, list[FormKitListValue]]


class FormKitSchemaAttributesCondition(BaseModel):
    if_: str = Field(alias="if")
    then_: FormKitAttributeValue = Field(alias="then")
    else_: FormKitAttributeValue | None = Field(alias="else")

    class Config:
        allow_population_by_field_name = True


class FormKitAttributeValue(BaseModel):
    """
    The possible value types of attributes (in the schema)
    """

    __root__: str | int | float | bool | None | FormKitSchemaAttributes | FormKitSchemaAttributesCondition


# ForwardRef allows for self referencing (recursive) models
FormKitSchemaAttributes = ForwardRef("FormKitSchemaAttributes")


class FormKitSchemaAttributes(BaseModel):
    __root__: dict[
        str,
        FormKitAttributeValue
        # | FormKitSchemaAttributes  # <-- Causes trouble for Ninja
        | FormKitSchemaAttributesCondition,
    ]


class FormKitSchemaProps(BaseModel):
    """
    Properties available in all schema nodes.
    """

    # "ForwardRefs" do not work well with django-ninja.
    # This would ideally be:
    # children: str | list[FormKitSchemaProps] | FormKitSchemaCondition | None = Field(
    #     default_factory=list
    # )
    children: str | list[FormKitSchemaProps | str] | FormKitSchemaConditionNoCircularRefs | None = Field()
    key: str | None
    if_condition: str | None = Field(alias="if")
    for_loop: FormKitListStatement | None = Field(alias="for")
    bind: str | None
    meta: FormKitSchemaMeta | None

    # These are not formal parts of spec, but
    # are attributes defined in ts as Record<string, any>
    # id: str | uuid.UUID | None = Field(None)
    html_id: str = Field(None, alias="id")
    name: str | None = Field(None)
    label: str | None = Field(None)
    help: str | None = Field(None)
    validation: str | None = Field(None)
    validationLabel: str | None = Field(None, alias="validation-label")
    validationVisibility: str | None = Field(None, alias="validation-visibility")
    validationMessages: str | dict[str, str] = Field(None, alias="validation-messages")
    placeholder: str | None = Field(None)
    value: str | None = Field(None)
    prefixIcon: str | None = Field(None, alias="prefix-icon")
    classes: dict[str, str] | None = Field(None)
    additional_props: dict[str, str] | None = Field(None)

    class Config:
        allow_population_by_field_name = True


class TextNode(FormKitSchemaProps):
    node_type: Literal["formkit"] = "formkit"
    formkit: Literal["text"] = "text"
    dollar_formkit: str = Field(default="text", alias="$formkit")
    text: str | None


class DateNode(FormKitSchemaProps):
    node_type: Literal["formkit"] = "formkit"
    formkit: Literal["date"] = "date"
    dollar_formkit: str = Field(default="date", alias="$formkit")


class CurrencyNode(FormKitSchemaProps):
    node_type: Literal["formkit"] = "formkit"
    formkit: Literal["currency"] = "currency"
    dollar_formkit: str = Field(default="currency", alias="$formkit")


class DatePickerNode(FormKitSchemaProps):
    node_type: Literal["formkit"] = "formkit"
    formkit: Literal["datepicker"] = "datepicker"
    dollar_formkit: str = Field(default="datepicker", alias="$formkit")
    calendarIcon: str = "calendar"
    format: str = "DD/MM/YY"
    nextIcon: str = "angleRight"
    prevIcon: str = "angleLeft"


class CheckBoxNode(FormKitSchemaProps):
    node_type: Literal["formkit"] = "formkit"
    formkit: Literal["checkbox"] = "checkbox"
    dollar_formkit: str = Field(default="checkbox", alias="$formkit")


class NumberNode(FormKitSchemaProps):
    node_type: Literal["formkit"] = "formkit"
    formkit: Literal["number"] = "number"
    dollar_formkit: str = Field(default="number", alias="$formkit")
    text: str | None
    max: int | None = None
    min: int | None = None
    step: str | None = None


class PasswordNode(FormKitSchemaProps):
    node_type: Literal["formkit"] = "formkit"
    formkit: Literal["password"] = "password"
    dollar_formkit: str = Field(default="password", alias="$formkit")
    name: str | None


class RadioNode(FormKitSchemaProps):
    node_type: Literal["formkit"] = "formkit"
    formkit: Literal["radio"] = "radio"
    dollar_formkit: str = Field(default="radio", alias="$formkit")
    name: str | None
    options: str | list[dict[str, Any]] | list[str] | dict[str, str] | None = Field(None)


class SelectNode(FormKitSchemaProps):
    node_type: Literal["formkit"] = "formkit"
    formkit: Literal["select"] = "select"
    dollar_formkit: str = Field(default="select", alias="$formkit")
    options: str | list[dict[str, Any]] | list[str] | dict[str, str] | None = Field(None)


class AutocompleteNode(FormKitSchemaProps):
    node_type: Literal["formkit"] = "formkit"
    formkit: Literal["autocomplete"] = "autocomplete"
    dollar_formkit: str = Field(default="autocomplete", alias="$formkit")
    options: str | list[dict[str, Any]] | list[str] | dict[str, str] | None = Field(None)


class EmailNode(FormKitSchemaProps):
    node_type: Literal["formkit"] = "formkit"
    formkit: Literal["email"] = "email"
    dollar_formkit: str = Field(default="email", alias="$formkit")


class TelNode(FormKitSchemaProps):
    node_type: Literal["formkit"] = "formkit"
    formkit: Literal["tel"] = "tel"
    dollar_formkit: str = Field(default="tel", alias="$formkit")


class DropDownNode(FormKitSchemaProps):
    node_type: Literal["formkit"] = "formkit"
    formkit: Literal["dropdown"] = "dropdown"
    dollar_formkit: str = Field(default="dropdown", alias="$formkit")
    options: str | list[dict[str, Any]] | list[str] | dict[str, str] | None = Field(None)
    empty_message: str | None = Field(None, alias="empty-message")
    select_icon: str | None = Field(None, alias="selectIcon")
    placeholder: str | None


class RepeaterNode(FormKitSchemaProps):
    node_type: Literal["formkit"] = "formkit"
    formkit: Literal["repeater"] = "repeater"
    dollar_formkit: str = Field(default="repeater", alias="$formkit")
    up_control: bool | None = Field(default=True, alias="upControl")
    down_control: bool | None = Field(default=True, alias="downControl")
    add_label: str | None = Field(default="Add another", alias="addLabel")
    name: str | None = None


class GroupNode(FormKitSchemaProps):
    node_type: Literal["formkit"] = "formkit"
    formkit: Literal["group"] = "group"
    dollar_formkit: str = Field(default="group", alias="$formkit")
    text: str | None


FormKitType = (
    TextNode
    | CheckBoxNode
    | PasswordNode
    | SelectNode
    | AutocompleteNode
    | EmailNode
    | NumberNode
    | RadioNode
    | GroupNode
    | DateNode
    | DatePickerNode
    | DropDownNode
    | RepeaterNode
    | TelNode
    | CurrencyNode
)

FormKitSchemaFormKit = Annotated[
    Union[
        TextNode,
        CheckBoxNode,
        PasswordNode,
        SelectNode,
        AutocompleteNode,
        EmailNode,
        NumberNode,
        RadioNode,
        GroupNode,
        DateNode,
        DatePickerNode,
        DropDownNode,
        RepeaterNode,
        TelNode,
        CurrencyNode,
    ],
    Field(discriminator="formkit"),
]


class FormKitSchemaDOMNode(FormKitSchemaProps):
    """
    HTML elements are defined using the $el property.
    You can use $el to render any HTML element.
    Attributes can be added with the attrs property,
    and content is assigned with the children property
    """

    node_type: Literal["element"] = "element"
    el: str = Field(alias="$el")
    attrs: FormKitSchemaAttributes | None

    class Config:
        allow_population_by_field_name = True


class FormKitSchemaComponent(FormKitSchemaProps):
    """
    Components can be defined with the $cmp property
    The $cmp property should be a string that references
    a globally defined component or a component passed
    into FormKitSchema with the library prop.
    """

    node_type: Literal["component"]

    cmp: str = Field(
        ...,
        alias="$cmp",
        description="The $cmp property should be a string that references a globally defined component or a component passed into FormKitSchema with the library prop.",  # noqa: E501
    )
    props: dict[str, str | Any] | None

    class Config:
        allow_population_by_field_name = True


# This necessary to properly "populate" some more complicated models
FormKitSchemaAttributesCondition.update_forward_refs()
FormKitAttributeValue.update_forward_refs()
FormKitSchemaDOMNode.update_forward_refs()
FormKitSchemaCondition.update_forward_refs()
FormKitSchemaComponent.update_forward_refs()


class FormKitTagParser(HTMLParser):
    """
    Reverse an HTML example to schema
    This is for lazy copy-pasting from the formkit website :)
    """

    def __init__(self, html_content, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.data: str | None = None

        self.current_tag: FormKitSchemaFormKit | None = None
        self.tags: list[FormKitSchemaFormKit] = []
        self.parents: list[FormKitSchemaFormKit] = []
        self.feed(html_content)

    def handle_starttag(self, tag, attrs):
        """
        Read anything that's a "formtag" type
        """
        if tag != "formkit":
            return
        props = dict(attrs)
        props["formkit"] = props.pop("type")

        tag = FormKitSchemaFormKit(**props)
        self.current_tag = tag

        if self.parents:
            self.parents[-1].children.append(tag)
        else:
            self.tags.append(tag)
            self.parents.append(tag)

    def handle_endtag(self, tag: str) -> None:
        if tag != "formkit":
            return
        if self.parents:
            self.parents.pop()

    def handle_data(self, data):
        if self.current_tag and data.strip():
            self.current_tag.children.append(data.strip())
            # Ensure that children is included even when "exclude_unset" is True
            # since we populated this after the initial tag build
            self.current_tag.__fields_set__.add("children")


FormKitSchemaDOMNode.update_forward_refs()

Model = TypeVar("Model", bound="BaseModel")
StrBytes = str | bytes

Node = Annotated[
    Union[
        FormKitSchemaFormKit,
        FormKitSchemaDOMNode,
        FormKitSchemaComponent,
        FormKitSchemaCondition,
    ],
    Field(discriminator="node_type"),
]

NODE_TYPE = Literal["condition", "formkit", "element", "component"]
FORMKIT_TYPE = Literal[
    "text",
    "tel",
    "currency",
    "select",
    "checkbox",
    "number",
    "group",
    "list",
    "password",
    "button",
    "radio",
    "form",
    "date",
    "datepicker",
    "dropdown",
    "repeater",
    "autocomplete",
    "email",
]


class Discriminators(TypedDict, total=False):
    node_type: NODE_TYPE
    formkit: FORMKIT_TYPE
    dollar_formkit: FORMKIT_TYPE


def get_node_type(obj: dict) -> Discriminators:
    """
    Pydantic requires nodes to be "differentiated" by a field value
    when used in a Union type situation.
    This function should return the 'node_type' values and if present 'Formkit' value
    which corresponds to the object being inspected.
    """
    fields: Discriminators = {}
    if "__root__" in obj:
        obj = obj["__root__"]
    for key, return_value in (
        # -- Loading from "raw json"
        ("$el", "element"),
        ("$formkit", "formkit"),
        ("dollar_formkit", "formkit"),
        ("$cmp", "component"),
        # -- Already in the database
        ("el", "element"),
        ("formkit", "formkit"),
        ("cmp", "component"),
        ("if", "condition"),
    ):
        if key in obj:
            fields["node_type"] = return_value
            # "formkit" and "$formkit" are aliases
            # Because we use these in a disriminated union, we can't alias
            # So for consistency: we always return both
            if key == "$formkit":
                fields["formkit"] = obj["$formkit"]
                fields["dollar_formkit"] = obj["$formkit"]
            elif key == "dollar_formkit":
                fields["formkit"] = obj["dollar_formkit"]
                fields["dollar_formkit"] = obj["dollar_formkit"]

            return fields
    raise KeyError("Could not determine node type")


class FormKitNode(BaseModel):
    __root__: Node

    @classmethod
    def parse_obj(cls: Type["Model"], obj: Any) -> "Model":
        """
        This classmethod differentiates between the different "Node" types
        when deserializing
        """
        if "id" in obj:
            obj["html_id"] = obj.pop("id")

        # Parsing is complicated by the fact that some instances use differentiation we
        # can't express in Python, so the `get_node_type` here tries to
        # translate the possible node types we received
        try:
            parsed = super().parse_obj({**get_node_type(obj), **obj})
        except KeyError as E:
            raise KeyError(f"Unable to parse content {obj} to a {cls}") from E

        # A FormKit node can have 'arbitrary' additional properties
        # For instance classes to apply to child nodes
        # here we can't realistically cover every scenario so
        # fall back to JSON storage for these
        if hasattr(parsed.__root__, "additional_props"):
            add_props_keys = (
                obj.keys()
                - parsed.__root__.dict().keys()
                - {"$formkit", "if", "for", "then", "else", "children", "dollar_formkit", "node_type", "formkit"}
            )
            for k in add_props_keys:
                parsed.__root__.additional_props = {k: obj[k]}

        # Recursively parse 'child' nodes back to Pydantic models for 'children'
        if getattr(parsed.__root__, "children", None):
            children = []
            for n in obj["children"]:
                # Especially for HTML, we can have a 'string' value
                # of HTML content
                if isinstance(n, str):
                    # What to do here? Append the `str` or ignore?
                    continue
                    # children.append(str)
                try:
                    children.append(cls.parse_obj(n).__root__)
                except Exception as E:
                    warnings.warn(f"{E}")
            parsed.__root__.children = children

        return parsed


class FormKitSchema(BaseModel):
    __root__: list[Node]

    @classmethod
    def parse_obj(cls: Type["Model"], obj: Any) -> "Model":
        return cls(__root__=[FormKitNode.parse_obj(_).__root__ for _ in obj])


FormKitSchema.update_forward_refs()
FormKitSchemaCondition.update_forward_refs()
PasswordNode.update_forward_refs()



from django.core.management.base import BaseCommand

from formkit_ninja import models
from formkit_ninja.management.commands.common_nodes import create_location_nodes


def create():
    municipality, admin_post, suco, aldeia = (node for (node, _) in create_location_nodes())
    locations, created = models.FormKitSchema.objects.get_or_create(key="SF 1.2 Location and Date")
    count, created = models.FormKitSchema.objects.get_or_create(key="SF 1.2 Community contribution")

    yield locations, created
    yield models.FormComponents.objects.update_or_create(
        defaults=dict(
            schema=locations,
            node=municipality,
        ),
        label="Municipalities for the SF 1.2 Location and Date schema",
    )

    yield models.FormComponents.objects.update_or_create(
        label="Administrative Posts for the SF 1.2 Location and Date schema",
        defaults=dict(
            schema=locations,
            node=admin_post,
        ),
    )

    yield models.FormComponents.objects.update_or_create(
        defaults=dict(
            schema=locations,
            node=suco,
        ),
        label="Suco for the SF 1.2 Location and Date schema",
    )

    yield models.FormComponents.objects.update_or_create(
        label="Aldeias for the SF 1.2 Location and Date schema",
        defaults=dict(
            schema=locations,
            node=aldeia,
        ),
    )

    date_component, created = models.FormKitSchemaNode.objects.update_or_create(
        label="SF 1.2 Date",
        defaults=dict(
            node_type="$formkit", node=dict(name="date", label="""$gettext("Date")""", calendarIcon="calendar")
        ),
    )

    yield date_component, created

    yield models.FormComponents.objects.update_or_create(
        label="Date picker for the SF 1.2 Location and Date schema",
        defaults=dict(
            schema=locations,
            node=date_component,
        ),
    )

    cc, cc_created = models.FormKitSchemaNode.objects.update_or_create(
        label="SF 1.2 Community Contribution",
        defaults=dict(
            node_type="$formkit",
            node=dict(
                label="Cost estimation",
                placeholder="Please enter",
                name="cost_estimation",
                min=0,
            ),
            additional_props={"icon": "las la-address-book"},
        ),
    )

    yield cc, cc_created

    yield models.FormComponents.objects.update_or_create(
        label="Community Contributions for SF 1.2",
        defaults=dict(
            schema=count,
            node=cc,
        ),
    )


class Command(BaseCommand):
    help = """
        Create the Partisipa 'SF12' form
        as a set of Django models / FormKit integration
    """

    def handle(self, *args, **options):
        for instance, created in create_location_nodes():
            if options["verbosity"] > 1:
                self.stdout.write(self.style.SUCCESS(f"{'Created' if created else 'Updated'} {instance}"))
        for instance, created in create():
            if options["verbosity"] > 1:
                self.stdout.write(self.style.SUCCESS(f"{'Created' if created else 'Updated'} {instance}"))



from enum import Enum

from django.core.management.base import BaseCommand

from formkit_ninja import models
from formkit_ninja.formkit_schema import FormKitNode


class TrainingOrMeeting(Enum):
    Formasaun = 1
    Enkontru = 2


class MeetingTypes(Enum):
    SucoSocialisation = 1
    AldeaSocialisation = 2
    CMTelectionatsuco = 4
    AnnualGrantAgreementSigned = 11
    ProjectImplementationagreementsigned = 16
    Accountabilitymeeting1 = 17
    Districtsocialization = 20
    Subdistrictsocialization = 21
    CMTnominationataldeia = 22
    Prioritysettingataldeia = 23
    Prioritysettingatsuco = 24
    Accountabilitymeeting2 = 28
    PrioritysettingataldeiaWomen = 40


def partisipants_schema():
    """
    Create the "Partisipants" section
    """
    partisipants, created = models.FormKitSchema.objects.get_or_create(key="SF 1.1 Partisipants")
    yield partisipants, created
    # This is a dump from python code
    partisipants_group = {
        "children": [
            {
                "key": "attendance_male",
                "if": "$get(activity_type).value !== '----' && $get(activity_subtype).value !== '----' && $get(activity_subtype).value != 16 && $get(activity_subtype).value != 40 && $get(activity_subtype).value != 11",
                "id": "attendance_male",
                "name": "attendance_male",
                "label": "Total participants male",
                "validation": "greaterThanOrEqualSum:kpa_male+community_member_male",
                "validation-messages": {
                    "greaterThanOrEqualSum": "The total participants male should be greater than or equal to the sum of Participants Suku Management Team (SMT) - male and Number of community members - male"
                },
                "node_type": "formkit",
                "formkit": "number",
                "$formkit": "number",
                "min": 0,
            },
            {
                "key": "attendance_female",
                "if": "$get(activity_type).value !== '----' && $get(activity_subtype).value !== '----' && $get(activity_subtype).value != 16 && $get(activity_subtype).value != 11",
                "id": "attendance_female",
                "name": "attendance_female",
                "label": "Total participants female",
                "validation": "greaterThanOrEqualSum:kpa_female+community_member_female",
                "validation-messages": {
                    "greaterThanOrEqualSum": "The total participants female should be greater than or equal to the sum of Participants Suku Management Team (SMT) - female and Number of community members - female"
                },
                "node_type": "formkit",
                "formkit": "number",
                "$formkit": "number",
                "min": 0,
            },
            {
                "key": "kpa_male",
                "if": "$get(activity_type).value !== '----' && $get(activity_subtype).value !== '----' && $get(activity_subtype).value != 16 && $get(activity_subtype).value != 20 && $get(activity_subtype).value != 21 && $get(activity_subtype).value != 40 && $get(activity_subtype).value != 11",
                "id": "kpa_male",
                "name": "kpa_male",
                "label": "Participants Suku Management Team (SMT) - male",
                "node_type": "formkit",
                "formkit": "number",
                "$formkit": "number",
                "min": 0,
            },
            {
                "key": "kpa_female",
                "if": "$get(activity_type).value !== '----' && $get(activity_subtype).value !== '----' && $get(activity_subtype).value != 16 && $get(activity_subtype).value != 11 && $get(activity_subtype).value != 20 && $get(activity_subtype).value != 21",
                "id": "kpa_female",
                "name": "kpa_female",
                "label": "Participants Suku Management Team (SMT) - female",
                "node_type": "formkit",
                "formkit": "number",
                "$formkit": "number",
                "min": 0,
            },
            {
                "key": "disable_male",
                "if": "$get(activity_type).value !== '----' && $get(activity_subtype).value !== '----' && $get(activity_subtype).value != 16 && $get(activity_subtype).value != 40 && $get(activity_subtype).value != 11",
                "id": "disable_male",
                "name": "disable_male",
                "label": "Number of People with Disability - male",
                "node_type": "formkit",
                "formkit": "number",
                "$formkit": "number",
                "min": 0,
            },
            {
                "key": "disable_female",
                "if": "$get(activity_type).value !== '----' && $get(activity_subtype).value !== '----' && $get(activity_subtype).value != 16 && $get(activity_subtype).value != 11",
                "id": "disable_female",
                "name": "disable_female",
                "label": "Number of People with Disability - female",
                "node_type": "formkit",
                "formkit": "number",
                "$formkit": "number",
                "min": 0,
            },
            {
                "key": "community_member_male",
                "if": "$get(activity_type).value !== '----' && $get(activity_subtype).value !== '----' && $get(activity_subtype).value != 16 && $get(activity_subtype).value != 20 && $get(activity_subtype).value != 21 && $get(activity_subtype).value != 40 && $get(activity_subtype).value != 11",
                "id": "community_member_male",
                "name": "community_member_male",
                "label": "Number of community members - male",
                "node_type": "formkit",
                "formkit": "number",
                "$formkit": "number",
                "min": 0,
            },
            {
                "key": "community_member_female",
                "if": "$get(activity_type).value !== '----' && $get(activity_subtype).value !== '----' && $get(activity_subtype).value != 16 && $get(activity_subtype).value != 11 && $get(activity_subtype).value != 20 && $get(activity_subtype).value != 21",
                "id": "community_member_female",
                "name": "community_member_female",
                "label": "Number of community members - female",
                "node_type": "formkit",
                "formkit": "number",
                "$formkit": "number",
                "min": 0,
            },
        ],
        "name": "partisipants",
        "node_type": "formkit",
        "formkit": "group",
        "$formkit": "group",
    }

    for node in partisipants_group["children"]:
        node_from_fk = FormKitNode.parse_obj(node).__root__
        schema_node, schema_node_created = models.FormKitSchemaNode.objects.update_or_create(
            label=f"SF 1.1 partisipants ({node_from_fk.name})",
            node_type="$formkit",
            defaults=dict(node=node_from_fk.dict()),
        )
        yield schema_node, schema_node_created
        component, component_created = models.FormComponents.objects.update_or_create(
            defaults=dict(
                schema=partisipants,
                node=schema_node,
            ),
            label=f"SF 1.1 partisipants ({node_from_fk.name})",
        )
        yield component, component_created


def meeting_information_schema():
    """
    Create or recreate the SF 1.1 Meeting Information schema
    """

    meeting_information, mi_created = models.FormKitSchema.objects.get_or_create(key="SF 1.1 Meeting Information")
    yield meeting_information, mi_created
    activity_type, activity_type_created = models.FormKitSchemaNode.objects.update_or_create(
        label="SF 1.1 Activity Type",
        node_type="$formkit",
        defaults=dict(
            node={
                "key": "activity_type",
                "html_id": "activity_type",
                "name": "activity_type",
                "label": "Meeting or Training",
                "node_type": "formkit",
                "formkit": "select",
            }
        ),
    )
    yield activity_type, activity_type_created

    yield models.Option.objects.update_or_create(
        field=activity_type, value=str(TrainingOrMeeting.Formasaun.value), label="Training"
    )
    yield models.Option.objects.update_or_create(
        field=activity_type, value=str(TrainingOrMeeting.Enkontru.value), label="Meeting"
    )

    yield models.FormComponents.objects.update_or_create(
        defaults=dict(
            schema=meeting_information,
            node=activity_type,
        ),
        label="SF 1.1 Meeting Information (activity type)",
    )

    # The JSON code for the Subtype nodes

    activity_subtype = FormKitNode.parse_obj(
        {
            "key": "activity_subtype",
            "if_condition": f"$get(activity_type).value == '{TrainingOrMeeting.Formasaun.value}'",
            "html_id": "activity_subtype",
            "name": "activity_subtype",
            "label": "Training Type",
            "node_type": "formkit",
            "formkit": "select",
            "dollar_formkit": "select",
        }
    ).__root__

    options_activity_subtype_training = [
        {"value": "----", "label": ""},
        {"value": "41", "label": "Module UKL - UKL Verification Guideline"},
        {"value": "39", "label": "Module L -Monitoring O & M"},
        {"value": "32", "label": "Module E -Proposal Writing"},
        {"value": "31", "label": "Module C -Sf Orientation"},
        {"value": "38", "label": "Module K -Financial Management 3"},
        {"value": "37", "label": "Module J -Operations and Maintainance"},
        {"value": "36", "label": "Module I -Monitoring & Accountability"},
        {"value": "35", "label": "Module H -Financial Management 2"},
        {"value": "3", "label": "Module D -Financial Management 1"},
        {"value": "30", "label": "Module B -Womens Empowerment 1"},
        {"value": "29", "label": "Module A -Introduction & Planning"},
        {"value": "34", "label": "Module G -Managing Construction 1"},
        {"value": "33", "label": "Module F -Womens Empowerment 2"},
    ]

    activity_subtype_meetings = FormKitNode.parse_obj(
        {
            "key": "activity_subtype",
            "if_condition": f"$get(activity_type).value == '{TrainingOrMeeting.Enkontru.value}'",
            "html_id": "activity_subtype",
            "name": "activity_subtype",
            "label": "Meeting Type",
            "node_type": "formkit",
            "formkit": "select",
            "dollar_formkit": "select",
        }
    ).__root__
    options_activity_subtype_meeting = [
        {"value": "----", "label": ""},
        {"value": "22", "label": "CMT nomination at aldeia"},
        {"value": "2", "label": "Aldea Socialisation"},
        {"value": "4", "label": "CMT election at suco"},
        {"value": "11", "label": "Annual Grant Agreement Signed"},
        {"value": "16", "label": "Project Implementation agreement signed"},
        {"value": "17", "label": "Accountability meeting 1"},
        {"value": "20", "label": "District socialization"},
        {"value": "21", "label": "Sub-district socialization"},
        {"value": "1", "label": "Suco Socialisation"},
        {"value": "23", "label": "Priority setting at aldeia"},
        {"value": "24", "label": "Priority setting at suco"},
        {"value": "28", "label": "Accountability meeting 2"},
        {"value": "40", "label": "Priority setting at aldeia-Women"},
    ]

    activity_subtype_node, c = models.FormKitSchemaNode.objects.update_or_create(
        label="SF 1.1 Activity Sub-Type (Training)",
        node_type="$formkit",
        defaults=dict(node=activity_subtype.dict(exclude_none=True)),
    )
    yield activity_subtype_node, c

    for option in options_activity_subtype_training:
        yield models.Option.objects.get_or_create(field=activity_subtype_node, **option)

    activity_subtype_node_meetings, c = models.FormKitSchemaNode.objects.update_or_create(
        label="SF 1.1 Activity Sub-Type (Meeting)",
        node_type="$formkit",
        defaults=dict(node=activity_subtype_meetings.dict(exclude_none=True)),
    )
    yield activity_subtype_node_meetings, c

    for option in options_activity_subtype_meeting:
        yield models.Option.objects.get_or_create(field=activity_subtype_node_meetings, **option)


def create_sf11_locations():
    """
    This is a special "SF11 Locations" schema with additional conditions
    For these 'SF11' type forms, we add some conditions as described in our spreadsheet
    """
    locations_sf11, c = models.FormKitSchema.objects.get_or_create(key="SF 1.1 Locations")
    yield locations_sf11, c
    district, c = models.FormKitSchemaNode.objects.update_or_create(
        label="municipality",
        defaults=dict(
            node_type="$formkit",
            node=dict(
                label="Municipality",
                html_id="municipality",
                name="municipality",
                key="municipality",
                formkit="select",
                options="$getLocations()",
            ),
        ),
    )
    yield district, c

    sf11_admin_post, c = models.FormKitSchemaNode.objects.get_or_create(
        label="Administrative Post select (SF 1.1 options)",
        defaults=dict(
            node_type="$formkit",
            node=dict(
                label="Administrative Post",
                html_id="admin_post",
                name="admin_post",
                key="admin_post",
                formkit="select",
                if_condition=" && ".join(
                    (
                        '$getLocations($get("municipality").value)',
                        f"get(activity_subtype).value !== '{MeetingTypes.Districtsocialization.value}'",
                    )
                ),
                options='$getLocations($get("municipality").value)',
            ),
        ),
    )
    yield sf11_admin_post, c

    sf11_suco, c = models.FormKitSchemaNode.objects.get_or_create(
        label="Suco select (SF 1.1 options)",
        defaults=dict(
            node_type="$formkit",
            node=dict(
                label="Suco",
                html_id="suco",
                name="suco",
                key="suco",
                formkit="select",
                if_condition=" && ".join(
                    (
                        '$getLocations($get("admin_post").value)',
                        f"get(activity_subtype).value !== '{MeetingTypes.Districtsocialization.value}'",
                        f"get(activity_subtype).value !== '{MeetingTypes.Subdistrictsocialization.value}'",
                    )
                ),
                options='$getLocations($get("municipality").value, $get("admin_post").value)',
            ),
        ),
    )
    yield sf11_suco, c

    sf11_aldeia, c = models.FormKitSchemaNode.objects.get_or_create(
        label="Aldeia select (SF 1.1 options)",
        defaults=dict(
            node_type="$formkit",
            node=dict(
                label="Aldeia",
                html_id="aldeia",
                name="aldeia",
                key="aldeia",
                formkit="select",
                if_condition=" && ".join(
                    [
                        '$getLocations($get("suco").value)',  # Has a suco value,
                        *[
                            f"get(activity_subtype).value !== '{item.value}'"
                            for item in [
                                MeetingTypes.Districtsocialization,
                                MeetingTypes.Subdistrictsocialization,
                                MeetingTypes.SucoSocialisation,
                                MeetingTypes.Prioritysettingatsuco,
                                MeetingTypes.CMTelectionatsuco,
                                MeetingTypes.AnnualGrantAgreementSigned,
                                MeetingTypes.ProjectImplementationagreementsigned,
                                MeetingTypes.Accountabilitymeeting1,
                                MeetingTypes.Accountabilitymeeting2,
                            ]
                        ],
                    ]
                ),
                options='$getLocations($get("municipality").value, $get("admin_post").value, $get("suco").value)',
            ),
        ),
    )
    yield sf11_aldeia, c

    yield models.FormComponents.objects.update_or_create(
        defaults=dict(
            schema=locations_sf11,
            node=district,
        ),
        label="Municipalities for the SF 1.1 Location Select schema",
    )

    yield models.FormComponents.objects.update_or_create(
        label="Administrative Posts for the SF 1.1 Location Select schema",
        defaults=dict(
            schema=locations_sf11,
            node=sf11_admin_post,
        ),
    )

    yield models.FormComponents.objects.update_or_create(
        defaults=dict(
            schema=locations_sf11,
            node=sf11_suco,
        ),
        label="Suco for the SF 1.1 Location Select schema",
    )

    yield models.FormComponents.objects.update_or_create(
        label="Aldeias for the for SF 1.1 Location Select schema",
        defaults=dict(
            schema=locations_sf11,
            node=sf11_aldeia,
        ),
    )


class Command(BaseCommand):
    help = """
        Create the Partisipa 'SF11' form
        as a set of Django models / FormKit integration
    """

    def handle(self, *args, **options):
        self.stdout.write(self.style.SUCCESS("Writing Partisipants schema"))
        for instance, created in partisipants_schema():
            if options["verbosity"] > 1:
                self.stdout.write(self.style.SUCCESS(f"{'Created' if created else 'Updated'} {instance}"))

        self.stdout.write(self.style.SUCCESS("Writing Meeting Information schema"))
        for instance, created in meeting_information_schema():
            if options["verbosity"] > 1:
                self.stdout.write(self.style.SUCCESS(f"{'Created' if created else 'Updated'} {instance}"))

        self.stdout.write(self.style.SUCCESS("Writing Locations schema for SF 11"))
        for instance, created in create_sf11_locations():
            if options["verbosity"] > 1:
                self.stdout.write(self.style.SUCCESS(f"{'Created' if created else 'Updated'} {instance}"))

        # Update / Create translated content hooks for "labels" and "options"
        for instance, created in models.FormKitSchemaNode.translatable_content():
            if options["verbosity"] > 1:
                self.stdout.write(self.style.SUCCESS(f"{'Created' if created else 'Updated'} {instance}"))
